/*
 * Copyright (C), 2000-2002 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>
#include <stdio.h>
#include <memory.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "net.h"
#include "ssl.h"


/* To be removed if included in the real monit!!!!!! */
#define error printf

/**
 *  SSL Socket methods.
 *
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */

/**
 * Embeds a socket in a ssl connection.
 * @param socket the socket to be used.
 * @return The ssl connection or NULL if an error occured.
 */

ssl_connection * embed_ssl_socket (int socket) {

  ssl_connection * ssl = new_ssl_connection();

  if ( socket >= 0 ) {
    
    ssl->socket= socket;

  } else {

    error("Socket error!");
    goto sslerror;
  }

  /* Init SSL stuff */
  SSLeay_add_ssl_algorithms();
  SSL_load_error_strings();

  if ((ssl->method= SSLv2_client_method()) == NULL ) {

    error("Cannot initilize SSL method!");
    goto sslerror;

  } 

  if ((ssl->ctx= SSL_CTX_new (ssl->method)) == NULL ) {

    error("Cannot initilize SSL server certificate handler!");
    goto sslerror;

  }

  if ((ssl->handler= SSL_new (ssl->ctx)) == NULL ) {

    error("Cannot initilize SSL handler!");
    goto sslerror;

  }

  SSL_set_fd (ssl->handler, ssl->socket);

  if (SSL_connect (ssl->handler) == -1) {
    
    error("Cannot connect with SSL service!");
    goto sslerror;

  }

  ssl->cipher= (char *) SSL_get_cipher(ssl->handler);

  if ((ssl->cert = SSL_get_peer_certificate (ssl->handler)) == NULL) {

    error("Cannot get SSL server certificate!");
    goto sslerror;

  } 

  ssl->cert_issuer=X509_NAME_oneline (X509_get_issuer_name (ssl->cert),0,0);
  ssl->cert_subject=X509_NAME_oneline (X509_get_subject_name (ssl->cert),0,0);

  return ssl;

 sslerror:

  cleanup_ssl_socket(ssl);
  return NULL;
} 

/**
 * Open a socket against hostname:port with the given protocol.
 * This socket is sent through a ssl connection.
 * The protocol is normaly either SOCK_STREAM or SOCK_DGRAM.
 * @param hostname The host to open a socket at
 * @param port The port number to connect to
 * @param protocol Socket protocol to use (SOCK_STREAM|SOCK_DGRAM)
 * @return The ssl connection or NULL if an error occured.
 */

ssl_connection * create_ssl_socket(char *hostname, int port, int protocol) { 

  int socket;

  if ((socket= create_socket(hostname, port, protocol)) == -1) {

    error("Cannot connect!");
    goto sslerror;

  }

  return embed_ssl_socket(socket);

 sslerror:

  return NULL;
}

/* Server code */

/**
 * Initilizes a ssl connection for server use.
 * @param pemfilename Filename for the key/cert file
 * @return An ssl connection, or NULL if an error occured.
 */

ssl_connection * init_ssl_server (char *pemfilename) {

  ssl_connection * ssl = new_ssl_connection();
  
  SSLeay_add_ssl_algorithms();
  SSL_load_error_strings();

  if ((ssl->method= SSLv23_server_method()) == NULL ) {

    error("Cannot initilize SSL method!");
    goto sslerror;

  }

  if ((ssl->ctx= SSL_CTX_new(ssl->method)) == NULL ) {

    error("Cannot initilize SSL server certificate handler!");
    goto sslerror;

  }

  if (SSL_CTX_use_certificate_file(ssl->ctx, pemfilename, SSL_FILETYPE_PEM) <= 0) {

    error("Cannot initilize SSL server certificate!");
    goto sslerror;

  }

  if (SSL_CTX_use_PrivateKey_file(ssl->ctx, pemfilename, SSL_FILETYPE_PEM) <= 0) {

    error("Cannot initilize SSL server private key!");
    goto sslerror;

  }

  if (!SSL_CTX_check_private_key(ssl->ctx)) {

    error("Private key does not match the certificate public key!");
    goto sslerror;

  }

  return ssl;

 sslerror:

  cleanup_ssl_socket(ssl);
  return NULL;

}

/**
 * Creates a server socket (SOCK_STREAM type) and binds it to the
 * specified local port number.  The socket get a ssl layer for
 * data transmission.
 * @param pemfilename Filename for the key/cert file
 * @param port The localhost port number to open
 * @param backlog The maximum queue length for incomming connections
 * @param bindAddr the local address the server will bind to
 * @return An ssl connection ready for accept, or NULL if an error occured.
 */

ssl_connection * create_ssl_server_socket(char * pemfilename, int port, int backlog, char *bindAddr) { 

  int socket;
  ssl_connection * ssl;

  if ((socket= create_server_socket(port, backlog, bindAddr)) == -1) {

    error("Cannot connect!");
    goto sslerror;

  }

  if (( ssl= init_ssl_server (pemfilename)) == NULL) {

    goto sslerror;

  }
  
  ssl->server_socket=socket;

  return ssl;

 sslerror:

  return NULL;
 
}

/**
 * Embeds an accepted server socket in an existing ssl connection.
 * @param ssl ssl connection
 * @param socket the socket to be used.
 * @return TRUE, or FALSE if an error has occured.
 */

int embed_accepted_ssl_socket(ssl_connection * ssl, int socket) {

  ssl->socket=socket;
 
  printf("start\n");
  if ((ssl->handler= SSL_new(ssl->ctx)) == NULL ) { 
 
    error("Cannot initilize SSL handler!"); 
 
  } 

  printf("handle\n");

  if ( socket >= 0 ) {
  
    SSL_set_fd(ssl->handler, ssl->socket);

  } else {

    error("Socket error!");
    goto sslerror;

  }

  printf("socket!\n");

  if (SSL_accept(ssl->handler) == -1) {

    error("Accept with SSL service has failed!");
    goto sslerror;
  }

  printf("accept!\n");

  ssl->cipher= (char *) SSL_get_cipher(ssl->handler);
  ssl->cert= SSL_get_peer_certificate(ssl->handler);

  if ( ssl->cert!= NULL ) {
    
    ssl->cert_issuer=X509_NAME_oneline(X509_get_issuer_name(ssl->cert), 0, 0);
    ssl->cert_subject=X509_NAME_oneline(X509_get_subject_name(ssl->cert), 0, 0);

  }

  return TRUE;

 sslerror:

  return FALSE;

}

/**
 * Do "accept" for a ssl server socket 
 * @param ssl ssl connection
 * @return TRUE, or FALSE if an error has occured.
 */

int accept_ssl_socket(ssl_connection * ssl) {

  int no_crypt_socket;
  int len= sizeof(struct sockaddr_in);
  struct sockaddr_in in;
 
  no_crypt_socket= accept(ssl->server_socket, (struct sockaddr*)&in, &len);

  printf("socket accept %i\n", no_crypt_socket);

  if ( no_crypt_socket >= 0 ) {

    return embed_accepted_ssl_socket(ssl, no_crypt_socket);

  } else {

    return FALSE;

  }

}

/* Generic code */

/**
 * Closes a ssl connection (ssl socket + net socket)
 * @param ssl ssl connection
 * @return TRUE, or FALSE if an error has occured.
 */

int close_ssl_socket(ssl_connection * ssl) {

  SSL_shutdown (ssl->handler);
  close(ssl->socket);
  cleanup_ssl_socket(ssl);

  return TRUE;
}

/**
 * Closes a ssl server connection (ssl socket + net socket)
 * @param ssl ssl connection
 * @return TRUE, or FALSE if an error has occured.
 */

int close_ssl_server_socket(ssl_connection * ssl) {

  close(ssl->server_socket);
  cleanup_ssl_socket(ssl);

  return TRUE;
}

/**
 * Garbage collection for non reusabe parts of the ssl connection
 * @param ssl ssl connection
 */

void cleanup_ssl_socket(ssl_connection *ssl) {

  if (ssl->cert != NULL) {

    X509_free(ssl->cert);
    ssl->cert= NULL;

  }

  if (ssl->handler != NULL) {
    
    SSL_free(ssl->handler);
    ssl->handler= NULL;

  }

  if (ssl->cert_issuer != NULL) {

    free(ssl->cert_issuer);
    ssl->cert_issuer= NULL;

  }

  if (ssl->cert_subject != NULL) {

    free(ssl->cert_subject);
    ssl->cert_subject= NULL;

  }

}

/**
 * Garbage collection for non-reusable parts a ssl connection
 * @param ssl ssl connection
 */

void delete_ssl_socket(ssl_connection *ssl) {

  cleanup_ssl_socket(ssl);

  if (ssl->ctx != NULL) {

    SSL_CTX_free(ssl->ctx);

  }

  free(ssl);

}


/**
 * Send data package though the ssl connection 
 * @param ssl ssl connection
 * @param buffer array containg the data
 * @param len size of the data container
 * @return TRUE, or FALSE if an error has occured.
 */

int send_ssl_socket(ssl_connection * ssl, void * buffer, int len) {

  return SSL_write (ssl->handler, (void *) buffer, len);

}

/**
 * Receive data package though the ssl connection 
 * @param ssl ssl connection
 * @param buffer array to hold the data
 * @param len size of the data container
 * @return TRUE, or FALSE if an error has occured.
 */

int recv_ssl_socket(ssl_connection * ssl, void * buffer, int len) {

  return SSL_read (ssl->handler, (void *) buffer, len);

}

/* ----------------------------------------------------------------- Private */

/**
 * Generate a new ssl connection
 * @return ssl connection container
 */

ssl_connection * new_ssl_connection() {

  ssl_connection * ssl = (ssl_connection *) NEW(ssl);

  ssl->handler= NULL;
  ssl->ctx= NULL;
  ssl->cert= NULL;
  ssl->method= NULL;
  ssl->cipher= NULL;
  ssl->socket= 0;

  return ssl;
}
